# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Dialog for warning when closing multiple documents."""

import aeidon
import gaupol

from aeidon.i18n   import _
from gi.repository import GObject
from gi.repository import Gtk

class MultiCloseDialog(Gtk.MessageDialog):

    """Dialog for warning when closing multiple documents."""

    def __init__(self, parent, application, pages):
        """Initialize a :class:`MultiCloseDialog` instance."""
        GObject.GObject.__init__(self,
                                 message_type=Gtk.MessageType.WARNING,
                                 text=_("Save changes to documents before closing?"),
                                 secondary_text=_("If you don't save, changes will be permanently lost."))

        self.application = application
        self.pages = tuple(pages)
        self._document_list = Gtk.ListBox()
        self._init_dialog(parent)
        self._init_document_list()

    def _init_dialog(self, parent):
        """Initialize the dialog."""
        self.add_button(_("Close _Without Saving"), Gtk.ResponseType.NO)
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Save"), Gtk.ResponseType.YES)
        self.set_default_response(Gtk.ResponseType.YES)
        self.set_transient_for(parent)
        self.set_modal(True)
        aeidon.util.connect(self, self, "response")

    def _init_document_list(self):
        """Initialize the list of unsaved documents."""
        self._document_list.set_selection_mode(Gtk.SelectionMode.NONE)
        self._document_list.set_show_separators(True)
        self._document_list.add_css_class("rich-list")
        for page in self.pages:
            if page.project.main_changed:
                self._append_document(page, aeidon.documents.MAIN)
            if page.project.tran_changed:
                self._append_document(page, aeidon.documents.TRAN)
        frame = Gtk.Frame(child=self._document_list)
        scroller = Gtk.ScrolledWindow(
            hscrollbar_policy=Gtk.PolicyType.NEVER,
            propagate_natural_width=True,
            propagate_natural_height=True)
        scroller.set_max_content_height(gaupol.util.lines_to_px(12))
        scroller.set_child(frame)
        scroller.set_visible(bool(self._get_check_buttons()))
        gaupol.util.pack_start_expand(self.get_message_area(), scroller)

    def _append_document(self, page, doc):
        """Append a check button row for `doc` of `page`."""
        check = Gtk.CheckButton(label=page.get_basename(doc),
                                active=True,
                                hexpand=True,
                                focusable=True)

        check.connect("toggled", self._on_check_button_toggled)
        check.gaupol_page = page
        check.gaupol_document = doc
        row = Gtk.ListBoxRow(activatable=False, selectable=False)
        row.set_child(check)
        self._document_list.append(row)

    @aeidon.deco.listify
    def _get_check_buttons(self):
        """Return all check buttons in the document list."""
        for i in range(1_000_000):
            row = self._document_list.get_row_at_index(i)
            if not row: break
            yield row.get_child()

    def _on_check_button_toggled(self, check_button):
        """Update save button sensitivity."""
        sensitive = any(x.get_active() for x in self._get_check_buttons())
        self.set_response_sensitive(Gtk.ResponseType.YES, sensitive)

    def _on_response(self, dialog, response):
        """Save the selected documents and close pages."""
        if response == Gtk.ResponseType.YES:
            for page in self.pages:
                self._save_and_close_page(page)
        if response == Gtk.ResponseType.NO:
            for page in self.pages:
                self.application.close(page, confirm=False)

    @aeidon.deco.silent(gaupol.Default)
    def _save_and_close_page(self, page):
        """Save the selected documents and close `page`."""
        for check in self._get_check_buttons():
            if check.gaupol_page is page and check.get_active():
                self.application.save(page, check.gaupol_document)
        self.application.close(page, confirm=False)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Entry for time data in format ``[-]HH:MM:SS.SSS``."""

import functools
import gaupol
import re

from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Gtk

def _blocked(function):
    """Decorator for methods to be run with buffer edits allowed."""
    @functools.wraps(function)
    def wrapper(entry, *args, **kwargs):
        entry.get_buffer().allow_edit = True
        try:
            return function(entry, *args, **kwargs)
        finally:
            entry.get_buffer().allow_edit = False
    return wrapper

class _TimeBuffer(Gtk.EntryBuffer):

    """
    Entry buffer that blocks all edits not explicitly allowed.

    Blocked insertions are routed to the entry's validating
    :meth:`TimeEntry._insert_text` instead; blocked deletions
    only move the cursor to the start of the range.
    """

    def __init__(self, entry):
        """Initialize a :class:`_TimeBuffer` instance."""
        GObject.GObject.__init__(self)
        self.allow_edit = False
        self._entry = entry

    def do_delete_text(self, position, n_chars):
        """Delete text if allowed, otherwise just move the cursor."""
        if self.allow_edit:
            return Gtk.EntryBuffer.do_delete_text(self, position, n_chars)
        gaupol.util.idle_add(self._entry.set_position, position)
        return 0

    def do_insert_text(self, position, chars, n_chars):
        """Insert text if allowed, otherwise route to validation."""
        if self.allow_edit:
            return Gtk.EntryBuffer.do_insert_text(self, position, chars, n_chars)
        gaupol.util.idle_add(self._entry._insert_text, chars)
        return 0

class TimeEntry(Gtk.Entry):

    """
    Entry for time data in format ``[-]HH:MM:SS.SSS``.

    This widget uses :func:`GLib.idle_add` a lot, which means that clients may
    need to call :func:`gaupol.util.iterate_main` to ensure proper updating.
    """
    _re_digit = re.compile(r"\d")
    _re_time = re.compile(r"^-?\d\d:[0-5]\d:[0-5]\d\.\d\d\d$")

    def __init__(self):
        """Initialize a :class:`TimeEntry` instance."""
        GObject.GObject.__init__(self)
        self.set_buffer(_TimeBuffer(self))
        self.set_width_chars(13)
        # Don't set max_length: the buffer clips insertions against it
        # before do_insert_text is called, so a full 13-character text
        # (any negative time) would block all typing. The buffer and
        # the validation regexp keep the length fixed anyway.
        self._init_signal_handlers()

    def _init_signal_handlers(self):
        """Initialize signal handlers."""
        text_widget = self.get_delegate()
        text_widget.connect("cut-clipboard", self._on_cut_clipboard)
        text_widget.connect("toggle-overwrite", self._on_toggle_overwrite)
        controller = Gtk.EventControllerKey()
        controller.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        controller.connect("key-pressed", self._on_key_pressed)
        self.add_controller(controller)

    @_blocked
    def _insert_text(self, value):
        """Insert `value` as text after validation."""
        pos = self.get_position()
        text = self.get_text()
        if pos == 0 and value.startswith("-"):
            text = text if text.startswith("-") else f"-{text}"
        length = len(value)
        text = text[:pos] + value + text[pos+length:]
        text = text.replace(",", ".")
        if not self._re_time.match(text): return
        self.set_text(text)
        self.set_position(pos)
        if length != 1: return
        self.set_position(pos + 1)
        if len(text) > pos + 1 and text[pos+1] in (":", "."):
            self.set_position(pos + 2)

    def _on_cut_clipboard(self, text_widget):
        """Change "cut-clipboard" signal to "copy-clipboard"."""
        text_widget.stop_emission_by_name("cut-clipboard")
        text_widget.emit("copy-clipboard")

    def _on_key_pressed(self, controller, keyval, keycode, state):
        """Change numbers to zero if Backspace or Delete pressed."""
        keys = (Gdk.KEY_BackSpace, Gdk.KEY_Delete)
        if not keyval in keys: return False
        if self.get_selection_bounds():
            gaupol.util.idle_add(self._zero_selection)
        elif keyval == Gdk.KEY_BackSpace:
            gaupol.util.idle_add(self._zero_previous)
        elif keyval == Gdk.KEY_Delete:
            gaupol.util.idle_add(self._zero_next)
        return True

    def _on_toggle_overwrite(self, text_widget):
        """Do not allow toggling overwrite."""
        text_widget.stop_emission_by_name("toggle-overwrite")

    @_blocked
    def _zero_next(self):
        """Change the next digit to zero."""
        pos = self.get_position()
        text = self.get_text()
        if pos >= len(text): return
        if pos == 0 and text.startswith("-"):
            self.set_text(text[1:])
            return self.set_position(0)
        if not text[pos].isdigit(): return
        self.set_text(text[:pos] + "0" + text[pos+1:])
        self.set_position(pos)

    @_blocked
    def _zero_previous(self):
        """Change the previous digit to zero."""
        pos = self.get_position()
        text = self.get_text()
        if pos <= 0: return
        if pos == 1 and text.startswith("-"):
            self.set_text(text[1:])
            return self.set_position(0)
        if not text[pos-1].isdigit():
            return self.set_position(pos-1)
        self.set_text(text[:pos-1] + "0" + text[pos:])
        self.set_position(pos-1)

    @_blocked
    def _zero_selection(self):
        """Change digits in selection to zero."""
        if not self.get_selection_bounds(): return
        a, z = self.get_selection_bounds()
        text = self.get_text()
        zero = self._re_digit.sub("0", text[a:z])
        self.set_text(text[:a] + zero + text[z:])
        self.set_position(a)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Text markup for the Sub Station Alpha format."""

import aeidon
import re

class SubStationAlpha(aeidon.Markup):

    """
    Text markup for the Sub Station Alpha format.

    Sub Station Alpha format contains a lot of markup tags of which the
    following are of interest to us. The generic reset ``{\\r}`` is used to
    revert to regular text, i.e. to close all open tags. Most of the tagging
    methods, e.g. :meth:`colorize`, leave tags unclosed, instead of explicitly
    closing them and possibly other tags with ``{\\r}``. Usually this is not a
    problem as such tags tend to be applied to the whole subtitle.

     * ``{\\b1}...........{\\b0}``
     * ``{\\i1}...........{\\i0}``
     * ``{\\fnNAME}............``
     * ``{\\fsPOINTS}..........``
     * ``{\\c&HBBGGRR&}........``
     * ``.................{\\r}``

     The hexadecimal color value is in reverse order, ``BBGGRR`` instead of the
     normal ``RRGGBB``. Furthermore, leading zeros can be omitted,
     e.g. ``ff00`` can be used instead of ``00ff00``.
    """

    _flags = re.DOTALL | re.MULTILINE | re.IGNORECASE
    format = aeidon.formats.SSA

    # Defined here so that AdvSubStationAlpha can override them.
    _closing_pattern = r"\{\\([bi])0\}"
    _opening_pattern = r"\{\\(?![bi]0)(b|i|c|fn|fs).*?\}"
    _reset_pattern   = r"\{\\r\}"

    def bolden(self, text, bounds=None):
        """Return bolded `text`."""
        a, z = bounds or (0, len(text))
        target = f"{{\\b1}}{text[a:z]}{{\\b0}}"
        return "".join((text[:a], target, text[z:]))

    def colorize(self, text, color, bounds=None):
        """Return `text` colorized to hexadecimal value."""
        a, z = bounds or (0, len(text))
        # Reverse the color value from RRGGBB to BBGGRR.
        color = f"{color[4:]}{color[2:4]}{color[:2]}"
        target = f"{{\\c&H{color}&}}{text[a:z]}"
        return "".join((text[:a], target, text[z:]))

    def fontify(self, text, font, bounds=None):
        """Return `text` changed to `font`."""
        a, z = bounds or (0, len(text))
        target = f"{{\\fn{font}}}{text[a:z]}"
        return "".join((text[:a], target, text[z:]))

    @property
    def italic_tag(self):
        """Regular expression for an italic markup tag."""
        return self._get_regex(r"\{\\i[01]\}")

    def italicize(self, text, bounds=None):
        """Return italicized `text`."""
        a, z = bounds or (0, len(text))
        target = f"{{\\i1}}{text[a:z]}{{\\i0}}"
        return "".join((text[:a], target, text[z:]))

    def _main_decode(self, text):
        """Return `text` with decodable markup decoded."""
        text = self._decode_b(text, r"\{\\b1\}(.*?)\{\\b[0\\]\}", 1)
        text = self._decode_c(text, r"\{\\c#(.+?)\}(.*?)\{\\c\\\}", 1, 2)
        text = self._decode_f(text, r"\{\\fn(.+?)\}(.*?)\{\\fn\\\}", 1, 2)
        text = self._decode_i(text, r"\{\\i1\}(.*?)\{\\i[0\\]\}", 1)
        text = self._decode_s(text, r"\{\\fs(\d+)\}(.*?)\{\\fs\\\}", 1, 2)
        return text

    def _post_decode(self, text):
        """Return `text` with markup finalized after decoding."""
        # Remove all unsupported markup tags.
        return self._substitute(text, r"\{\\.*?\}", "")

    def _pre_decode(self, text):
        """Return `text` with markup prepared for decoding."""
        text = self._pre_decode_break(text)
        text = self._pre_decode_reset(text)
        text = self._pre_decode_color(text)
        return text

    def _pre_decode_break(self, text):
        """
        Return `text` with combined markup tags separated.

        For example, ``{\\b1\\i1}`` is replaced with ``{\\b1}{\\i1}``.
        """
        parts = text.split("\\")
        for i in range(1, len(parts)):
            text_so_far = "\\".join(parts[:i])
            if text_so_far.endswith("{"): continue
            opening_index = text_so_far.rfind("{")
            closing_index = text_so_far.rfind("}")
            if opening_index > closing_index:
                parts[i - 1] += "}{"
        return "\\".join(parts)

    def _pre_decode_color(self, text):
        """
        Return `text` with colors converted to standard hexadecimal form.

        Color tags are converted from ``{\\c&HBBGGRR&}`` to ``{\\c#RRGGBB}``.
        """
        pattern = r"\{\\c&H([0-9a-fA-F]*)&\}"
        regex = self._get_regex(pattern)
        match = regex.search(text)
        if match is None: return text
        color = f"{match.group(1):0>6s}".replace(" ", "0")
        color = f"{color[4:]}{color[2:4]}{color[:2]}"
        text = regex.sub(rf"{{\\c#{color}}}", text, 1)
        return self._pre_decode_color(text)

    def _pre_decode_reset(self, text):
        """
        Return `text` with all markup tags closed explicitly.

        Tags of form ``{\\nameVALUE}`` are closed with ``{\\name\\}``.
        The returned text will not contain reset ``{\\r}`` tags.
        """
        re_opening = self._get_regex(self._opening_pattern)
        re_closing = self._get_regex(self._closing_pattern)
        re_reset = self._get_regex(self._reset_pattern)
        parts = re_reset.split(text + "{\\r}")
        for i, part in enumerate(parts):
            opening_matches = [x for x in re_opening.finditer(part)]
            closing_matches = [x for x in re_closing.finditer(part)]
            # Find out which tags have already been closed.
            for j in reversed(range(len(closing_matches))):
                closing_core = closing_matches[j].group(1)
                for k in range(len(opening_matches)):
                    opening_core = opening_matches[k].group(1)
                    if opening_core == closing_core:
                        opening_matches.pop(k)
                        break
            # Add artificial closing tags to close remaining tags.
            for j in reversed(range(len(opening_matches))):
                core = opening_matches[j].group(1)
                parts[i] += f"{{\\{core}\\}}"
        return "".join(parts)

    def scale(self, text, size, bounds=None):
        """Return `text` scaled to `size`."""
        a, z = bounds or (0, len(text))
        target = f"{{\\fs{size}}}{text[a:z]}"
        return "".join((text[:a], target, text[z:]))

    @property
    def tag(self):
        """Regular expression for any markup tag."""
        return self._get_regex(r"\{\\.*?\}")
